/* 
 
  iPodME [iPod Media Encoder]
  ---------------------------
  version 1.1, January 10th, 2008

  Copyright (C) 2008 Noda (Lasorsa Yohan)

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:

  1. The origin of this software must not be misrepresented; you must not
     claim that you wrote the original software. If you use this software
     in a product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.

*/

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.Diagnostics;
using System.IO;

namespace iPodME
{
    public partial class MainWindow : Form
    {
        #region listbox data & values

        public static string[] priority_text = 
        {
            "Idle",
            "Normal",
            "High",
        };

        public static ProcessPriorityClass[] priority_val =
        {
            ProcessPriorityClass.Idle,
            ProcessPriorityClass.Normal,
            ProcessPriorityClass.AboveNormal
        };

        public static string[] size_text =
        {
            "640x480",
            "480x320",
            "320x240"
        };

        public static string[] encoder_text =
        {
            "x264",
            "xvid",
        };

        public static string[] enc_mode_text = 
        {
            "Average bitrate",
            "Constant quality",
        };

        public static string[] enc_mode_val = 
        {
            "-b ",
            "-crf ",
            "-b ",
            "-qscale "
        };

        public static string[] aud_freq_text = 
        {
            "22050",
            "44100",
            "48000",
        };

        public static string[] aud_br_text = 
        {
            "64",
            "96",
            "128",
            "160"
        };

        public static string[] profile_text = 
        {
            "Slow, size",
            "Slow, quality",
            "Fast, size",
            "Fast, quality",
            "Turbo, size",
            "Turbo, quality",
            "Custom",
        };

        public static string[] enc_profile_text = 
        {
            "Default",
            "Good quality",
            "Best quality"
        };

        #endregion

        #region encoding profiles & parameters

        private const string defaultEncOpts = "-maxrate 1500k -bufsize 4M -g 250 -coder 0 -threads auto -acodec libfaac -ac 2";

        public static string[] enc_profile_val = 
        {
            "",
            "-subq 4 -refs 2 -me hex -partitions +parti4x4+partp4x4+partp8x8+partb8x8 ",
            "-subq 5 -me umh -me_range 16 -trellis 2 -mbd 2 -refs 3 -partitions +parti4x4+partp4x4+partp8x8+partb8x8 ",
            "",
            "-trellis 2 -cmp 2 -subcmp 2 ",
            "-me umh -me_range 16 -trellis 2 -cmp 2 -subcmp 2 ",
        };

        public static string[] profile_val = 
        {
            "-vcodec libx264 -level 30 -crf 24 -ab 128k -subq 5 -me umh -me_range 16 -trellis 1 -mbd 2 -refs 3 -partitions +parti4x4+partp4x4+partp8x8+partb8x8",
            "-vcodec libx264 -level 30 -crf 19 -ab 128k -subq 5 -me umh -me_range 16 -trellis 1 -mbd 2 -refs 3 -partitions +parti4x4+partp4x4+partp8x8+partb8x8",
            "-vcodec libx264 -level 30 -crf 24 -ab 128k -subq 4 -refs 2 -me hex -partitions +parti4x4+partp4x4+partp8x8+partb8x8",
            "-vcodec libx264 -level 30 -crf 19 -ab 128k -subq 4 -refs 2 -me hex -partitions +parti4x4+partp4x4+partp8x8+partb8x8",
            "-vcodec libxvid -qscale 6 -ab 128k",
            "-vcodec libxvid -qscale 4 -ab 128k",
            ""
        };

        #endregion

        #region constants

        public static Size window_std = new Size(304, 354);
        public static Size window_co = new Size(304, 480);
        public static Size window_cus_std = new Size(500, 354);
        public static Size window_cus_co = new Size(500, 480);
        public static Size console_std = new Size(288, 120);
        public static Size console_cus = new Size(484, 120);

        private const string prog_info = "iPodME v1.1 by Noda // 2008";
        private const string encName = "\\ffmpeg.exe";
        private const string subfName = "\\vsfilter.dll";

        #endregion

        private string tempPath = Path.GetTempPath() + Path.GetRandomFileName();
        private List<string> files = new List<string>();
        private Process proc;
        private bool cancel = false, getInfos, subtitles;
        private int count, curr_duration, curr_width, curr_height;
        private string curr_file = "";
        private int[] quantizer = { 21, 3 };
        private int bitrate = 512;


        public MainWindow()
        {
            // init window
            InitializeComponent();

            sizeSelect.DataSource = size_text;
            prioSelect.DataSource = priority_text;
            profileSelect.DataSource = profile_text;
            encoderSelect.DataSource = encoder_text;
            encProfSelect.DataSource = enc_profile_text;
            encModeSelect.DataSource = enc_mode_text;
            audFreqSelect.DataSource = aud_freq_text;
            audRateSelect.DataSource = aud_br_text;

            profileSelect.SelectedIndex = 2;
            sizeSelect.SelectedIndex = 1;
            encProfSelect.SelectedIndex = 1;
            encModeSelect.SelectedIndex = 1;
            audFreqSelect.SelectedIndex = 1;
            audRateSelect.SelectedIndex = 2;

            consoleTextBox.AppendText("\n\r");
            CheckForIllegalCrossThreadCalls = false;
            setProgress(prog_info);
            Size = window_std;
            groupBox2.Visible = false;
            groupBox3.Visible = false;
            groupBox4.Visible = false;
            groupBox5.Visible = false;

            // create the temporary files
            try
            {
                // create ffmpeg exe
                Directory.CreateDirectory(tempPath);
                Stream file = new FileStream(tempPath + encName, FileMode.Create, FileAccess.Write);
                BinaryWriter writer = new BinaryWriter(file);
                writer.Write(iPodME.Properties.Resources.ffmpeg);
                writer.Close();
                file.Close();
                file.Dispose();

                // create vsfilter dll
                file = new FileStream(tempPath + subfName, FileMode.Create, FileAccess.Write);
                writer = new BinaryWriter(file);
                writer.Write(iPodME.Properties.Resources.VSFilter);
                writer.Close();
                file.Close();
                file.Dispose();
            }
            catch (Exception)
            {
                MessageBox.Show("Error creating temp file.");
                cleanQuit();
                Dispose();
            }
        }

        private void MainWindow_ClientSizeChanged(object sender, EventArgs e)
        {
            if (WindowState == FormWindowState.Minimized)
            {
                ShowInTaskbar = false;
                notifyIcon.Visible = true;
            }
            else
            {
                ShowInTaskbar = true;
                notifyIcon.Visible = false;
            }
        }

        private void notifyIcon_MouseClick(object sender, MouseEventArgs e)
        {
            if (WindowState != FormWindowState.Minimized)
            {
                Activate();
            }
            else
            {
                WindowState = FormWindowState.Normal;
            }
        }

        private void addFiles_Click(object sender, EventArgs e)
        {
            openFileDialog.ShowDialog();
        }

        private void openFileDialog_FileOk(object sender, CancelEventArgs e)
        {
            if (openFileDialog.FileNames != null)
            {
                // add files
                foreach (string s in openFileDialog.FileNames)
                    files.Add(s);
                listBox.DataSource = null;
                listBox.DataSource = files;
            }
        }

        private void listBox_DragDrop(object sender, DragEventArgs e)
        {
            if (e.Data.GetDataPresent(DataFormats.FileDrop))
            {
                string[] dragFiles = (string[])e.Data.GetData(DataFormats.FileDrop);
                // add files
                foreach (string s in dragFiles)
                    files.Add(s);
                listBox.DataSource = null;
                listBox.DataSource = files;
            }
        }

        private void listBox_DragEnter(object sender, DragEventArgs e)
        {
            if (e.Data.GetDataPresent(DataFormats.FileDrop))
                e.Effect = DragDropEffects.Copy;
            else
                e.Effect = DragDropEffects.None;
        }

        private void convert_Click(object sender, EventArgs e)
        {
            if (backgroundWorker.IsBusy)
            {
                try
                {
                    cancel = true;
                    consoleTextBox.AppendText("Conversion cancelled.\n\r");
                    if (proc != null && !proc.HasExited)
                        proc.Kill();
                }
                catch (Exception)
                { }
            }
            else
            {
                convert.Text = "Cancel";
                profileSelect.Enabled = false;
                sizeSelect.Enabled = false;
                stretchCheck.Enabled = false;
                subCheck.Enabled = false;
                groupBox3.Enabled = false;
                groupBox4.Enabled = false;
                groupBox5.Enabled = false;
                backgroundWorker.RunWorkerAsync();
            }
        }

        private void initEncoderProc()
        {
            // init ffmpeg process
            proc = new System.Diagnostics.Process();
            proc.StartInfo.FileName = tempPath + encName;
            proc.StartInfo.UseShellExecute = false;
            proc.StartInfo.RedirectStandardError = true;
            proc.StartInfo.RedirectStandardOutput = true;
            proc.StartInfo.CreateNoWindow = true;

            proc.ErrorDataReceived += new System.Diagnostics.DataReceivedEventHandler(proc_DataReceived);
            proc.OutputDataReceived += new System.Diagnostics.DataReceivedEventHandler(proc_DataReceived);
        }

        private void backgroundWorker_DoWork(object sender, DoWorkEventArgs e)
        {
            count = 1;

            // encode files
            while (files.Count > 0)
            {
                string filename = files[0];
                curr_file = filename;
                string args_info = "-y -i \"" + filename + "\"";

                // initialize some values
                curr_duration = 0;
                curr_height = 0;
                curr_width = 0;
                getInfos = true;
                subtitles = false;

                initEncoderProc();
                proc.StartInfo.Arguments = args_info;

                try
                {
                    setProgress("Initializing encoder...");

                    // first get infos from file
                    proc.Start();
                    proc.PriorityClass = priority_val[prioSelect.SelectedIndex];
                    proc.BeginErrorReadLine();
                    proc.BeginOutputReadLine();
                    proc.WaitForExit();
                    proc.CancelErrorRead();
                    proc.CancelOutputRead();
                    proc.Close();
                    proc.Dispose();

                    // then convert it
                    filename = checkAndInitSubs(filename);
                    string args_conv = computeArgs(filename, computeSize());

                    consoleTextBox.AppendText("ffmpeg.exe " + args_conv + "\n\r");
                    getInfos = false;
                    initEncoderProc();
                    proc.StartInfo.Arguments = args_conv;

                    proc.Start();
                    proc.PriorityClass = priority_val[prioSelect.SelectedIndex];
                    proc.BeginErrorReadLine();
                    proc.BeginOutputReadLine();
                    proc.WaitForExit();
                    proc.CancelErrorRead();
                    proc.CancelOutputRead();
                    proc.Close();
                    proc.Dispose();
                    count++;

                    // remove AVS script if subtitles were used
                    if (subtitles)
                        File.Delete(filename);

                    if (!cancel)
                    {
                        // conversion finished
                        files.Remove(curr_file);
                        listBox.DataSource = null;
                        listBox.DataSource = files;
                    }
                    else
                    {
                        break;
                    }
                }
                catch (Exception ex)
                {
                    MessageBox.Show(ex.Message);
                    break;
                }
            }
        }

        private void backgroundWorker_RunWorkerCompleted(object sender, RunWorkerCompletedEventArgs e)
        {
            proc = null;
            curr_file = "";
            convert.Text = "Convert";
            profileSelect.Enabled = true;
            sizeSelect.Enabled = true;
            stretchCheck.Enabled = true;
            subCheck.Enabled = true;
            groupBox3.Enabled = true;
            groupBox4.Enabled = true;
            groupBox5.Enabled = true;
            cancel = false;
            setProgress(prog_info);
            consoleTextBox.AppendText("Ready.\n\r");
        }

        private void proc_DataReceived(object sender, DataReceivedEventArgs e)
        {
            try
            {
                if (getInfos)
                {
                    if ((curr_duration == 0) && e.Data.Contains("Duration:"))
                        setDuration(e.Data);
                    else if ((curr_height == 0) && (curr_width == 0) && e.Data.Contains("Video:"))
                        setSize(e.Data);
                }
                else
                {
                    if (e.Data.StartsWith("frame="))
                        updateProgress(e.Data);
                    else
                        consoleTextBox.AppendText(e.Data + "\r");
                }
            }
            catch (Exception)
            { }
        }

        private void cleanQuit()
        {
            try
            {
                cancel = true;
                if (proc != null && !proc.HasExited)
                {
                    proc.Kill();
                    proc.WaitForExit();
                }
                notifyIcon.Visible = false;
                notifyIcon.Dispose();
                File.Delete(tempPath + encName);
                File.Delete(tempPath + subfName);
                Directory.Delete(tempPath);
            }
            catch (Exception)
            { }
        }

        private void MainWindow_FormClosing(object sender, FormClosingEventArgs e)
        {
            if (proc != null && !proc.HasExited)
            {
                // confirm exit
                if (MessageBox.Show("Do you really want to quit?\nCurrent conversion will be cancelled!", "Confirm exit", MessageBoxButtons.YesNo) == DialogResult.No)
                {
                    e.Cancel = true;
                    return;
                }
            }
            cleanQuit();
        }

        private void prioSelect_SelectedIndexChanged(object sender, EventArgs e)
        {
            if (proc != null && !proc.HasExited)
            {
                proc.PriorityClass = priority_val[prioSelect.SelectedIndex];
            }
        }

        private void clear_Click(object sender, EventArgs e)
        {
            files.Clear();
            if (curr_file != "")
                files.Add(curr_file);
            listBox.DataSource = null;
            listBox.DataSource = files;
        }

        private void showConsole_CheckedChanged(object sender, EventArgs e)
        {
            resizeWindow();
        }

        private void profileSelect_SelectedIndexChanged(object sender, EventArgs e)
        {
            resizeWindow();
        }

        private void encModeSelect_SelectedIndexChanged(object sender, EventArgs e)
        {
            if (encModeSelect.SelectedIndex == 0)
            {
                rateControlLabel.Text = "Bitrate (kbps) :";
                rateBox.Text = bitrate.ToString();
            }
            else
            {
                rateControlLabel.Text = "Quantizer :";
                rateBox.Text = quantizer[encoderSelect.SelectedIndex].ToString();
            }
        }

        private void rateBox_Validated(object sender, EventArgs e)
        {
            if (encModeSelect.SelectedIndex == 0)
            {
                try
                {
                    int num = int.Parse(rateBox.Text);
                    if (num >= 100 && num <= 1500)
                    {
                        bitrate = num;
                    }
                    else if (num > 1500)
                    {
                        bitrate = 1500;
                        rateBox.Text = bitrate.ToString();
                    }
                    else if (num < 100)
                    {
                        bitrate = 100;
                        rateBox.Text = bitrate.ToString();
                    }
                }
                catch (Exception)
                {
                    rateBox.Text = bitrate.ToString();
                }
            }
            else
            {
                try
                {
                    int num = int.Parse(rateBox.Text);
                    if (num >= 1 && num <= 51)
                    {
                        quantizer[encoderSelect.SelectedIndex] = num;
                    }
                    else if (num > 51)
                    {
                        quantizer[encoderSelect.SelectedIndex] = 51;
                        rateBox.Text = quantizer[encoderSelect.SelectedIndex].ToString();
                    }
                    else if (num < 1)
                    {
                        quantizer[encoderSelect.SelectedIndex] = 1;
                        rateBox.Text = quantizer[encoderSelect.SelectedIndex].ToString();
                    }
                }
                catch (Exception)
                {
                    rateBox.Text = quantizer[encoderSelect.SelectedIndex].ToString();
                }
            }
        }

        private void consoleTextBox_TextChanged(object sender, EventArgs e)
        {
            consoleTextBox.ScrollToCaret();
        }

        private void resampleCheck_CheckedChanged(object sender, EventArgs e)
        {
            if (resampleCheck.Checked)
            {
                audFreqLabel.Enabled = true;
                audFreqSelect.Enabled = true;
            }
            else
            {
                audFreqLabel.Enabled = false;
                audFreqSelect.Enabled = false;
            }
        }

        private void listBox_DataSourceChanged(object sender, EventArgs e)
        {
            if (files.Count == 0)
                convert.Enabled = false;
            else
                convert.Enabled = true;
        }

        private void encoderSelect_SelectedIndexChanged(object sender, EventArgs e)
        {
            if (encModeSelect.SelectedIndex == 0)
                rateBox.Text = bitrate.ToString();
            else
                rateBox.Text = quantizer[encoderSelect.SelectedIndex].ToString();
        }

        private void resizeWindow()
        {
            if (showConsole.Checked)
            {
                if (profileSelect.SelectedIndex == 6)
                {
                    Size = window_cus_co;
                    groupBox3.Visible = true;
                    groupBox4.Visible = true;
                    groupBox5.Visible = true;
                    groupBox2.Size = console_cus;
                }
                else
                {
                    groupBox2.Size = console_std;
                    groupBox3.Visible = false;
                    groupBox4.Visible = false;
                    groupBox5.Visible = false;
                    Size = window_co;
                }
                groupBox2.Visible = true;
            }
            else
            {
                if (profileSelect.SelectedIndex == 6)
                {
                    Size = window_cus_std;
                    groupBox3.Visible = true;
                    groupBox4.Visible = true;
                    groupBox5.Visible = true;
                    groupBox2.Size = console_cus;
                }
                else
                {
                    groupBox2.Size = console_std;
                    groupBox3.Visible = false;
                    groupBox4.Visible = false;
                    groupBox5.Visible = false;
                    Size = window_std;
                }
                groupBox2.Visible = false;
            }
        }

        private void setProgress(string s)
        {
            notifyIcon.Text = s;
            infoBox.Text = s;
        }

        private void setDuration(string s)
        {
            try
            {
                string dur = s.Substring(s.IndexOf("Duration:") + 10, 8);
                curr_duration = int.Parse(dur.Substring(0, 2)) * 3600 + int.Parse(dur.Substring(3, 2)) * 60 + int.Parse(dur.Substring(6, 2));
            }
            catch(Exception) 
            { }        
        }

        private void setSize(string s)
        {
            try
            {
                int idx_st = s.IndexOf(',', s.IndexOf(',') + 1) + 2;
                string siz = s.Substring(idx_st, s.IndexOf(' ', idx_st) - idx_st);
                int idx_ax = siz.IndexOf('x') + 1;
                curr_width = int.Parse(siz.Substring(0, idx_ax - 1));
                curr_height = int.Parse(siz.Substring(idx_ax, siz.Length - idx_ax));
            }
            catch (Exception)
            { }
        }
        
        private void updateProgress(string s)
        {
            try
            {
                string fps = s.Substring(s.IndexOf("fps=") + 4, 3);
                int idx_at = s.IndexOf("time=") + 5;
                int time = int.Parse(s.Substring(idx_at, s.IndexOf('.', idx_at) - idx_at));
                setProgress("Converting file " + count + "/" + (files.Count + count - 1) + "    fps:" + fps + "    progress: " + (time * 100 / curr_duration) + "%");
            }
            catch (Exception)
            { }
        }

        private string computeSize()
        {
            string size = "";

            if (stretchCheck.Checked)
            {
                size = " -s " + size_text[sizeSelect.SelectedIndex];
            }
            else
            {
                string newsize = size_text[sizeSelect.SelectedIndex];
                int idx_ax = newsize.IndexOf('x') + 1;
                int width = int.Parse(newsize.Substring(0, idx_ax - 1));
                int height = int.Parse(newsize.Substring(idx_ax, newsize.Length - idx_ax));
                float ar = (float)curr_width / (float)curr_height;
                int nwidth = width, nheight;

                // check if resize is needed
                if (curr_width > width || curr_height > height)
                {
                    // first try to fit width
                    nheight = (int)Math.Round((float)width / ar, 0);
                    nheight = (int)(16 * Math.Round((float)nheight / 16, 0));  // round to nearest multple of 16

                    // check if the new height fits in the desired one
                    if (nheight > height)
                    {
                        // if not, fit the height
                        nheight = height;
                        nwidth = (int)Math.Round((float)height * ar, 0);
                        nwidth = (int)(16 * Math.Round((float)nwidth / 16, 0));  // round to nearest multple of 16
                    }

                    size = " -s " + nwidth + "x" + nheight;
                }
            }            
            return size;
        }

        private string computeArgs(string filename, string size)
        {
            string args = "";

            args += "-y -i \"" + filename + "\" -f mp4 ";

            if (profileSelect.SelectedIndex == 6)
            {
                args += "-vcodec lib" + encoderSelect.Text + " ";
                args += enc_profile_val[encProfSelect.SelectedIndex + encoderSelect.SelectedIndex * 3];
                args += enc_mode_val[encModeSelect.SelectedIndex + encoderSelect.SelectedIndex * 2];
                if (encModeSelect.SelectedIndex == 0) args += bitrate.ToString() + "k -bt " + bitrate.ToString() + "k ";
                else args += quantizer[encoderSelect.SelectedIndex].ToString() + " ";
                if (deinterlaceCheck.Checked) args += "-deinterlace ";
                args += defaultEncOpts + " ";
                if (resampleCheck.Checked) args += "-ar " + audFreqSelect.Text + " ";
                args += "-ab " + audRateSelect.Text + "k";
                args += size + " ";
                args += addParamBox.Text;
                args += " \"" + Directory.GetCurrentDirectory() + "\\" + Path.GetFileNameWithoutExtension(filename) + ".mp4\"";
            }
            else
            {
                args += profile_val[profileSelect.SelectedIndex] + size + " " + defaultEncOpts + " -ar 44100 \"" + Directory.GetCurrentDirectory() + "\\" + Path.GetFileNameWithoutExtension(filename) + ".mp4\"";
            }

            return args;
        }

        private string checkAndInitSubs(string filename)
        {
            string filepath_noext = Path.GetDirectoryName(filename) + "\\" + Path.GetFileNameWithoutExtension(filename);

            if (subCheck.Checked && File.Exists(filepath_noext + ".srt"))
            {
                string vsf_path = tempPath + subfName;
                subtitles = true;
                
                // check is VobSub is already installed
                if (File.Exists(Environment.GetFolderPath(Environment.SpecialFolder.System) + "\\vsfilter.dll"))
                    vsf_path = Environment.GetFolderPath(Environment.SpecialFolder.System) + "\\vsfilter.dll";

                // create AVS script
                Stream file = new FileStream(tempPath + "\\" + Path.GetFileNameWithoutExtension(filename) + ".avs", FileMode.Create, FileAccess.Write);
                StreamWriter writer = new StreamWriter(file);
                writer.Write("DirectShowSource(\"" + filename + "\")\r\n");
                writer.Write("LoadPlugin(\"" + vsf_path + "\")\r\n");
                writer.Write("TextSub(\"" + filepath_noext + ".srt" + "\")\r\n");
                writer.Close();
                file.Close();
                file.Dispose();

                return (tempPath + "\\" + Path.GetFileNameWithoutExtension(filename) + ".avs");
            }

            return filename;
        }

    }
}